import functools
import math
import numbers

from . import islhelper

from .islhelper import mainctx, libisl
from .geometry import GeometricObject, Point, Vector
from .linexprs import Expression, Symbol, Rational
from .domains import Domain


__all__ = [
    'Polyhedron',
    'Lt', 'Le', 'Eq', 'Ne', 'Ge', 'Gt',
    'Empty', 'Universe',
]


class Polyhedron(Domain):

    __slots__ = (
        '_equalities',
        '_inequalities',
        '_constraints',
        '_symbols',
        '_dimension',
    )

    def __new__(cls, equalities=None, inequalities=None):
        if isinstance(equalities, str):
            if inequalities is not None:
                raise TypeError('too many arguments')
            return cls.fromstring(equalities)
        elif isinstance(equalities, GeometricObject):
            if inequalities is not None:
                raise TypeError('too many arguments')
            return equalities.aspolyhedron()
        if equalities is None:
            equalities = []
        else:
            for i, equality in enumerate(equalities):
                if not isinstance(equality, Expression):
                    raise TypeError('equalities must be linear expressions')
                equalities[i] = equality.scaleint()
        if inequalities is None:
            inequalities = []
        else:
            for i, inequality in enumerate(inequalities):
                if not isinstance(inequality, Expression):
                    raise TypeError('inequalities must be linear expressions')
                inequalities[i] = inequality.scaleint()
        symbols = cls._xsymbols(equalities + inequalities)
        islbset = cls._toislbasicset(equalities, inequalities, symbols)
        return cls._fromislbasicset(islbset, symbols)

    @property
    def equalities(self):
        return self._equalities

    @property
    def inequalities(self):
        return self._inequalities

    @property
    def constraints(self):
        return self._constraints

    @property
    def polyhedra(self):
        return self,

    def disjoint(self):
        """
        Return this set as disjoint.
        """
        return self

    def isuniverse(self):
        """
        Return true if this set is the Universe set.
        """
        islbset = self._toislbasicset(self.equalities, self.inequalities,
            self.symbols)
        universe = bool(libisl.isl_basic_set_is_universe(islbset))
        libisl.isl_basic_set_free(islbset)
        return universe

    def aspolyhedron(self):
        """
        Return polyhedral hull of this set.
        """
        return self

    def __contains__(self, point):
        if not isinstance(point, Point):
            raise TypeError('point must be a Point instance')
        if self.symbols != point.symbols:
            raise ValueError('arguments must belong to the same space')
        for equality in self.equalities:
            if equality.subs(point.coordinates()) != 0:
                return False
        for inequality in self.inequalities:
            if inequality.subs(point.coordinates()) < 0:
                return False
        return True

    def subs(self, symbol, expression=None):
        equalities = [equality.subs(symbol, expression)
            for equality in self.equalities]
        inequalities = [inequality.subs(symbol, expression)
            for inequality in self.inequalities]
        return Polyhedron(equalities, inequalities)

    @classmethod
    def _fromislbasicset(cls, islbset, symbols):
        islconstraints = islhelper.isl_basic_set_constraints(islbset)
        equalities = []
        inequalities = []
        for islconstraint in islconstraints:
            constant = libisl.isl_constraint_get_constant_val(islconstraint)
            constant = islhelper.isl_val_to_int(constant)
            coefficients = {}
            for index, symbol in enumerate(symbols):
                coefficient = libisl.isl_constraint_get_coefficient_val(islconstraint,
                    libisl.isl_dim_set, index)
                coefficient = islhelper.isl_val_to_int(coefficient)
                if coefficient != 0:
                    coefficients[symbol] = coefficient
            expression = Expression(coefficients, constant)
            if libisl.isl_constraint_is_equality(islconstraint):
                equalities.append(expression)
            else:
                inequalities.append(expression)
        libisl.isl_basic_set_free(islbset)
        self = object().__new__(Polyhedron)
        self._equalities = tuple(equalities)
        self._inequalities = tuple(inequalities)
        self._constraints = tuple(equalities + inequalities)
        self._symbols = cls._xsymbols(self._constraints)
        self._dimension = len(self._symbols)
        return self

    @classmethod
    def _toislbasicset(cls, equalities, inequalities, symbols):
        dimension = len(symbols)
        indices = {symbol: index for index, symbol in enumerate(symbols)}
        islsp = libisl.isl_space_set_alloc(mainctx, 0, dimension)
        islbset = libisl.isl_basic_set_universe(libisl.isl_space_copy(islsp))
        islls = libisl.isl_local_space_from_space(islsp)
        for equality in equalities:
            isleq = libisl.isl_equality_alloc(libisl.isl_local_space_copy(islls))
            for symbol, coefficient in equality.coefficients():
                islval = str(coefficient).encode()
                islval = libisl.isl_val_read_from_str(mainctx, islval)
                index = indices[symbol]
                isleq = libisl.isl_constraint_set_coefficient_val(isleq,
                    libisl.isl_dim_set, index, islval)
            if equality.constant != 0:
                islval = str(equality.constant).encode()
                islval = libisl.isl_val_read_from_str(mainctx, islval)
                isleq = libisl.isl_constraint_set_constant_val(isleq, islval)
            islbset = libisl.isl_basic_set_add_constraint(islbset, isleq)
        for inequality in inequalities:
            islin = libisl.isl_inequality_alloc(libisl.isl_local_space_copy(islls))
            for symbol, coefficient in inequality.coefficients():
                islval = str(coefficient).encode()
                islval = libisl.isl_val_read_from_str(mainctx, islval)
                index = indices[symbol]
                islin = libisl.isl_constraint_set_coefficient_val(islin,
                    libisl.isl_dim_set, index, islval)
            if inequality.constant != 0:
                islval = str(inequality.constant).encode()
                islval = libisl.isl_val_read_from_str(mainctx, islval)
                islin = libisl.isl_constraint_set_constant_val(islin, islval)
            islbset = libisl.isl_basic_set_add_constraint(islbset, islin)
        return islbset

    @classmethod
    def fromstring(cls, string):
        domain = Domain.fromstring(string)
        if not isinstance(domain, Polyhedron):
            raise ValueError('non-polyhedral expression: {!r}'.format(string))
        return domain

    def __repr__(self):
        if self.isempty():
            return 'Empty'
        elif self.isuniverse():
            return 'Universe'
        else:
            strings = []
            for equality in self.equalities:
                strings.append('Eq({}, 0)'.format(equality))
            for inequality in self.inequalities:
                strings.append('Ge({}, 0)'.format(inequality))
            if len(strings) == 1:
                return strings[0]
            else:
                return 'And({})'.format(', '.join(strings))

    def _repr_latex_(self):
        if self.isempty():
            return '$\\emptyset$'
        elif self.isuniverse():
            return '$\\Omega$'
        else:
            strings = []
            for equality in self.equalities:
                strings.append('{} = 0'.format(equality._repr_latex_().strip('$')))
            for inequality in self.inequalities:
                strings.append('{} \\ge 0'.format(inequality._repr_latex_().strip('$')))
            return '${}$'.format(' \\wedge '.join(strings))

    @classmethod
    def fromsympy(cls, expr):
        domain = Domain.fromsympy(expr)
        if not isinstance(domain, Polyhedron):
            raise ValueError('non-polyhedral expression: {!r}'.format(expr))
        return domain

    def tosympy(self):
        import sympy
        constraints = []
        for equality in self.equalities:
            constraints.append(sympy.Eq(equality.tosympy(), 0))
        for inequality in self.inequalities:
            constraints.append(sympy.Ge(inequality.tosympy(), 0))
        return sympy.And(*constraints)

    @classmethod
    def _polygon_inner_point(cls, points):
        symbols = points[0].symbols
        coordinates = {symbol: 0 for symbol in symbols}
        for point in points:
            for symbol, coordinate in point.coordinates():
                coordinates[symbol] += coordinate
        for symbol in symbols:
            coordinates[symbol] /= len(points)
        return Point(coordinates)

    @classmethod
    def _sort_polygon_2d(cls, points):
        if len(points) <= 3:
            return points
        o = cls._polygon_inner_point(points)
        angles = {}
        for m in points:
            om = Vector(o, m)
            dx, dy = (coordinate for symbol, coordinate in om.coordinates())
            angle = math.atan2(dy, dx)
            angles[m] = angle
        return sorted(points, key=angles.get)

    @classmethod
    def _sort_polygon_3d(cls, points):
        if len(points) <= 3:
            return points
        o = cls._polygon_inner_point(points)
        a = points[0]
        oa = Vector(o, a)
        norm_oa = oa.norm()
        for b in points[1:]:
            ob = Vector(o, b)
            u = oa.cross(ob)
            if not u.isnull():
                u = u.asunit()
                break
        else:
            raise ValueError('degenerate polygon')
        angles = {a: 0.}
        for m in points[1:]:
            om = Vector(o, m)
            normprod = norm_oa * om.norm()
            cosinus = max(oa.dot(om) / normprod, -1.)
            sinus = u.dot(oa.cross(om)) / normprod
            angle = math.acos(cosinus)
            angle = math.copysign(angle, sinus)
            angles[m] = angle
        return sorted(points, key=angles.get)

    def faces(self):
        vertices = self.vertices()
        faces = []
        for constraint in self.constraints:
            face = []
            for vertex in vertices:
                if constraint.subs(vertex.coordinates()) == 0:
                    face.append(vertex)
            faces.append(face)
        return faces

    def plot(self):
        """
        Display 3D plot of set. 
        """
        import matplotlib.pyplot as plt
        import matplotlib.patches as patches

        if len(self.symbols)> 3:
            raise TypeError

        elif len(self.symbols) == 2:
            import pylab
            points = []  
            for verts in self.vertices():
                    pairs=()
                    for coordinate, point in verts.coordinates():
                        pairs = pairs + (float(point),)
                    points.append(pairs)
            cent=(sum([p[0] for p in points])/len(points),sum([p[1] for p in points])/len(points))
            points.sort(key=lambda p: math.atan2(p[1]-cent[1],p[0]-cent[0]))
            pylab.scatter([p[0] for p in points],[p[1] for p in points])
            pylab.gca().add_patch(patches.Polygon(points,closed=True,fill=True))
            pylab.grid()
            pylab.show()

        elif len(self.symbols)==3:
            from mpl_toolkits.mplot3d import Axes3D
            from mpl_toolkits.mplot3d.art3d import Poly3DCollection
            faces = self.faces()
            fig = plt.figure()
            ax = Axes3D(fig)
            for face in faces:
                points = []
                vertices = Polyhedron._sort_polygon_3d(face)
                for verts in vertices:
                    pairs=()
                    for coordinate, point in verts.coordinates():
                        pairs = pairs + (float(point),)
                    points.append(pairs)
                collection = Poly3DCollection([points], alpha=0.7)
                face_color = [0.5, 0.5, 1] # alternative: matplotlib.colors.rgb2hex([0.5, 0.5, 1])
                collection.set_facecolor(face_color)
                ax.add_collection3d(collection)
            ax.set_xlabel('X')   
            ax.set_xlim(0, 5)
            ax.set_ylabel('Y')
            ax.set_ylim(0, 5)
            ax.set_zlabel('Z')
            ax.set_zlim(0, 5)
            plt.grid()      
            plt.show()
        return points
    
    @classmethod
    def limit(cls, faces, variable, lim):
        sym = []
        if variable is 'x':
            n = 0
        elif variable is 'y':
            n = 1
        elif variable is 'z':
            n = 2
        for face in faces:
            for vert in face:
                coordinates = vert.coordinates()
                for point in enumerate(coordinates):
                        coordinates.get(n)
                        sym.append(points)
        if lim == 0:
            value = min(sym)
        else:
            value = max(sym)
        return value

def _polymorphic(func):
    @functools.wraps(func)
    def wrapper(left, right):
        if not isinstance(left, Expression):
            if isinstance(left, numbers.Rational):
                left = Rational(left)
            else:
                raise TypeError('left must be a a rational number '
                    'or a linear expression')
        if not isinstance(right, Expression):
            if isinstance(right, numbers.Rational):
                right = Rational(right)
            else:
                raise TypeError('right must be a a rational number '
                    'or a linear expression')
        return func(left, right)
    return wrapper

@_polymorphic
def Lt(left, right):
    """
    Return true if the first set is less than the second.
    """
    return Polyhedron([], [right - left - 1])

@_polymorphic
def Le(left, right):
    """
    Return true the first set is less than or equal to the second.
    """
    return Polyhedron([], [right - left])

@_polymorphic
def Eq(left, right):
    """
    Return true if the sets are equal.
    """
    return Polyhedron([left - right], [])

@_polymorphic
def Ne(left, right):
    """
    Return true if the sets are NOT equal.
    """
    return ~Eq(left, right)

@_polymorphic
def Gt(left, right):
    """
    Return true if the first set is greater than the second set.
    """
    return Polyhedron([], [left - right - 1])

@_polymorphic
def Ge(left, right):
    """
    Return true if the first set is greater than or equal the second set.
    """
    return Polyhedron([], [left - right])


Empty = Eq(1, 0)

Universe = Polyhedron([])
